==> core-impl-192.7142.79-sources.jar: com/intellij/ide/plugins/LoadDescriptorsContext.java <==
// Copyright 2000-2019 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.ide.plugins;

import com.intellij.concurrency.SameThreadExecutorService;
import com.intellij.openapi.util.SafeJdomFactory;
import com.intellij.util.SmartList;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.intellij.util.containers.ContainerUtil;
import com.intellij.util.containers.HashSetInterner;
import com.intellij.util.containers.Interner;
import org.jdom.*;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.concurrent.ExecutorService;

final class LoadDescriptorsContext implements AutoCloseable {
  @NotNull
  private final ExecutorService myExecutorService;

  private final Collection<Interner<String>> myInterners;

  // synchronization will ruin parallel loading, so, string pool is local per thread
  private final ThreadLocal<SafeJdomFactory> myThreadLocalXmlFactory;

  LoadDescriptorsContext(boolean isParallel) {
    int maxThreads = isParallel ? (Runtime.getRuntime().availableProcessors() - 1) : 1;
    if (maxThreads > 1) {
      myExecutorService = AppExecutorUtil.createBoundedApplicationPoolExecutor("PluginManager Loader", maxThreads);
      myInterners = Collections.newSetFromMap(ContainerUtil.newConcurrentMap(maxThreads));
    }
    else {
      myExecutorService = new SameThreadExecutorService();
      myInterners = new SmartList<>();
    }

    myThreadLocalXmlFactory = ThreadLocal.withInitial(() -> {
      PluginXmlFactory factory = new PluginXmlFactory();
      myInterners.add(factory.stringInterner);
      return factory;
    });
  }

  @NotNull
  ExecutorService getExecutorService() {
    return myExecutorService;
  }

  @Nullable
  public SafeJdomFactory getXmlFactory() {
    return myThreadLocalXmlFactory.get();
  }

  @Override
  public void close() {
    if (myExecutorService instanceof SameThreadExecutorService) {
      myThreadLocalXmlFactory.remove();
      return;
    }

    myExecutorService.submit(() -> {
      for (Interner<String> interner : myInterners) {
        interner.clear();
      }
    });
    myExecutorService.shutdown();
  }

  /**
   * Consider using some threshold in StringInterner - CDATA is not interned at all,
   * but maybe some long text for Text node doesn't make sense to intern too.
   */
  // don't intern CDATA - in most cases it is used for some unique large text (e.g. plugin description)
  private final static class PluginXmlFactory extends SafeJdomFactory.BaseSafeJdomFactory {
    // doesn't make sense to intern class name since it is unique
    // ouch, do we really cannot agree how to name implementation class attribute?
    private static final List<String> CLASS_NAME_LIST = Arrays.asList(
      "implementation-class", "implementation",
      "serviceImplementation", "class", "className", "beanClass",
      "serviceInterface", "interface", "interfaceClass", "instance",
      "qualifiedName");

    private static final Set<String> CLASS_NAMES = ContainerUtil.newIdentityTroveSet(CLASS_NAME_LIST);

    private final Interner<String>
      stringInterner = new HashSetInterner<String>(ContainerUtil.concat(CLASS_NAME_LIST, IdeaPluginDescriptorImpl.SERVICE_QUALIFIED_ELEMENT_NAMES)) {
      @NotNull
      @Override
      public String intern(@NotNull String name) {
        // doesn't make any sense to intern long texts (JdomInternFactory doesn't intern CDATA, but plugin description can be simply Text)
        return name.length() < 64 ? super.intern(name) : name;
      }
    };

    @NotNull
    @Override
    public Interner<String> stringInterner() {
      return stringInterner;
    }

    @NotNull
    @Override
    public Element element(@NotNull String name, @Nullable Namespace namespace) {
      return super.element(stringInterner.intern(name), namespace);
    }

    @NotNull
    @Override
    public Attribute attribute(@NotNull String name, @NotNull String value, @Nullable AttributeType type, @Nullable Namespace namespace) {
      String internedName = stringInterner.intern(name);
      if (CLASS_NAMES.contains(internedName)) {
        return super.attribute(internedName, value, type, namespace);
      }
      else {
        return super.attribute(internedName, stringInterner.intern(value), type, namespace);
      }
    }

    @NotNull
    @Override
    public Text text(@NotNull String text, @NotNull Element parentElement) {
      if (CLASS_NAMES.contains(parentElement.getName())) {
        return super.text(text, parentElement);
      }
      else {
        return super.text(stringInterner.intern(text), parentElement);
      }
    }
  }
}

==> core-impl-193.6494.30-sources.jar: com/intellij/ide/plugins/LoadDescriptorsContext.java <==
// Copyright 2000-2019 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.ide.plugins;

import com.intellij.openapi.util.SafeJdomFactory;
import com.intellij.util.ConcurrencyUtil;
import com.intellij.util.SmartList;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.intellij.util.containers.ContainerUtil;
import com.intellij.util.containers.HashSetInterner;
import com.intellij.util.containers.Interner;
import org.jdom.*;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.concurrent.ExecutorService;

final class LoadDescriptorsContext implements AutoCloseable {
  @NotNull
  private final ExecutorService myExecutorService;

  private final Collection<Interner<String>> myInterners;

  // synchronization will ruin parallel loading, so, string pool is local per thread
  private final ThreadLocal<SafeJdomFactory> myThreadLocalXmlFactory;
  private final int myMaxThreads;

  LoadDescriptorsContext(boolean isParallel) {
    myMaxThreads = isParallel ? (Runtime.getRuntime().availableProcessors() - 1) : 1;
    if (myMaxThreads > 1) {
      myExecutorService = AppExecutorUtil.createBoundedApplicationPoolExecutor("PluginManager Loader", myMaxThreads, false);
      myInterners = Collections.newSetFromMap(ContainerUtil.newConcurrentMap(myMaxThreads));
    }
    else {
      myExecutorService = ConcurrencyUtil.newSameThreadExecutorService();
      myInterners = new SmartList<>();
    }

    myThreadLocalXmlFactory = ThreadLocal.withInitial(() -> {
      PluginXmlFactory factory = new PluginXmlFactory();
      myInterners.add(factory.stringInterner);
      return factory;
    });
  }

  @NotNull
  ExecutorService getExecutorService() {
    return myExecutorService;
  }

  @Nullable
  public SafeJdomFactory getXmlFactory() {
    return myThreadLocalXmlFactory.get();
  }

  @Override
  public void close() {
    if (myMaxThreads <= 1) {
      myThreadLocalXmlFactory.remove();
      return;
    }

    myExecutorService.execute(() -> {
      for (Interner<String> interner : myInterners) {
        interner.clear();
      }
    });
    myExecutorService.shutdown();
  }

  /**
   * Consider using some threshold in StringInterner - CDATA is not interned at all,
   * but maybe some long text for Text node doesn't make sense to intern too.
   */
  // don't intern CDATA - in most cases it is used for some unique large text (e.g. plugin description)
  private final static class PluginXmlFactory extends SafeJdomFactory.BaseSafeJdomFactory {
    // doesn't make sense to intern class name since it is unique
    // ouch, do we really cannot agree how to name implementation class attribute?
    private static final List<String> CLASS_NAME_LIST = Arrays.asList(
      "implementation-class", "implementation",
      "serviceImplementation", "class", "className", "beanClass",
      "serviceInterface", "interface", "interfaceClass", "instance",
      "qualifiedName");

    private static final Set<String> CLASS_NAMES = ContainerUtil.newIdentityTroveSet(CLASS_NAME_LIST);

    private final Interner<String>
      stringInterner = new HashSetInterner<String>(ContainerUtil.concat(CLASS_NAME_LIST, IdeaPluginDescriptorImpl.SERVICE_QUALIFIED_ELEMENT_NAMES)) {
      @NotNull
      @Override
      public String intern(@NotNull String name) {
        // doesn't make any sense to intern long texts (JdomInternFactory doesn't intern CDATA, but plugin description can be simply Text)
        return name.length() < 64 ? super.intern(name) : name;
      }
    };

    @NotNull
    @Override
    public Interner<String> stringInterner() {
      return stringInterner;
    }

    @NotNull
    @Override
    public Element element(@NotNull String name, @Nullable Namespace namespace) {
      return super.element(stringInterner.intern(name), namespace);
    }

    @NotNull
    @Override
    public Attribute attribute(@NotNull String name, @NotNull String value, @Nullable AttributeType type, @Nullable Namespace namespace) {
      String internedName = stringInterner.intern(name);
      if (CLASS_NAMES.contains(internedName)) {
        return super.attribute(internedName, value, type, namespace);
      }
      else {
        return super.attribute(internedName, stringInterner.intern(value), type, namespace);
      }
    }

    @NotNull
    @Override
    public Text text(@NotNull String text, @NotNull Element parentElement) {
      if (CLASS_NAMES.contains(parentElement.getName())) {
        return super.text(text, parentElement);
      }
      else {
        return super.text(stringInterner.intern(text), parentElement);
      }
    }
  }
}
